# -*- coding: utf-8 -*-
"""
Created on Fri Mar 27 13:11:26 2015

@author: mcgibbon
"""
from __future__ import division, absolute_import, unicode_literals
import numpy as np
import re
import six

derivative_prog = re.compile(r'd(.+)d(p|x|y|theta|z|sigma|t|lat|lon)')
from textwrap import wrap


def quantity_string(name, quantity_dict):
    '''Takes in an abbreviation for a quantity and a quantity dictionary,
       and returns a more descriptive string of the quantity as "name (units)."
       Raises ValueError if the name is not in quantity_dict
    '''
    if name not in quantity_dict.keys():
        raise ValueError('{0} is not present in quantity_dict'.format(name))
    return '{0} ({1})'.format(quantity_dict[name]['name'],
                              quantity_dict[name]['units'])


def strings_to_list_string(strings):
    '''Takes a list of strings presumably containing words and phrases,
       and returns a "list" form of those strings, like:

       >>> strings_to_list_string(('cats', 'dogs'))
       >>> 'cats and dogs'

       or

       >>> strings_to_list_string(('pizza', 'pop', 'chips'))
       >>> 'pizza, pop, and chips'

       Raises ValueError if strings is empty.
    '''
    if isinstance(strings, six.string_types):
        raise TypeError('strings must be an iterable of strings, not a string '
                        'itself')
    if len(strings) == 0:
        raise ValueError('strings may not be empty')
    elif len(strings) == 1:
        return strings[0]
    elif len(strings) == 2:
        return ' and '.join(strings)
    else:
        return '{0}, and {1}'.format(', '.join(strings[:-1]),
                                     strings[-1])


def assumption_list_string(assumptions, assumption_dict):
    '''Takes in a list of short forms of assumptions and an assumption
       dictionary, and returns a "list" form of the long form of the
       assumptions.

       Raises ValueError if one of the assumptions is not in assumption_dict.
    '''
    if isinstance(assumptions, six.string_types):
        raise TypeError('assumptions must be an iterable of strings, not a '
                        'string itself')
    for a in assumptions:
        if a not in assumption_dict.keys():
            raise ValueError('{} not present in assumption_dict'.format(a))
    assumption_strings = [assumption_dict[a] for a in assumptions]
    return strings_to_list_string(assumption_strings)


def quantity_spec_string(name, quantity_dict):
    '''Returns a quantity specification for docstrings. Example:
       >>> quantity_spec_string('Tv')
       >>> 'Tv : float or ndarray\n    Data for virtual temperature.'
    '''
    if name not in quantity_dict.keys():
        raise ValueError('{0} not present in quantity_dict'.format(name))
    s = '{0} : float or ndarray\n'.format(name)
    s += doc_paragraph('Data for {0}.'.format(
        quantity_string(name, quantity_dict)), indent=4)
    return s


def doc_paragraph(s, indent=0):
    '''Takes in a string without wrapping corresponding to a paragraph,
       and returns a version of that string wrapped to be at most 80
       characters in length on each line.
       If indent is given, ensures each line is indented to that number
       of spaces.
    '''
    return '\n'.join([' '*indent + l for l in wrap(s, width=80-indent)])


def parse_derivative_string(string, quantity_dict):
    '''
    Assuming the string is of the form d(var1)d(var2), returns var1, var2.
    Raises ValueError if the string is not of this form, or if the vars
    are not keys in the quantity_dict, or if var2 is not a coordinate-like
    variable.
    '''
    match = derivative_prog.match(string)
    if match is None:
        raise ValueError('string is not in the form of a derivative')
    varname = match.group(1)
    coordname = match.group(2)
    if (varname not in quantity_dict.keys() or
            coordname not in quantity_dict.keys()):
        raise ValueError('variable in string not a valid quantity')
    return varname, coordname


def landsea_mask(lat, lon, basemap=None, basemap_lat=None, basemap_lon=None):
    '''
    Calculates a land sea mask for a given latitude and longitude array.

    Parameters
    ----------
    lat : ndarray
        Latitudes in degrees N.
    lon : ndarray
        Longitudes in degrees E.
    basemap: ndarray, optional
        A 2D array of type byte or int that contains the baseline land sea
        mask from which the new land sea mask is generated. By default uses
        the grid distributed with NCL, available at
        https://www.ncl.ucar.edu/Applications/Data/cdf/landsea.nc
    basemap_lat: ndarray, optional
        The latitudes of basemap. If not specified, assumes a regularly spaced
        grid from -90 to 90 degrees.
    basemap_lon: ndarray, optional
        The longitudes of basemap. If not specified, assumes a regularly spaced
        grid from -180 to 180 degrees.
    '''
    raise NotImplementedError


def gaussian_latitudes(nlat):
    '''
    Generates gaussian latitudes.

    Parameters
    ----------
    nlat : int
        The number of latitudes desired

    Returns
    -------
    lat : ndarray
        A one-dimensional array containing the gaussian latitudes.
    '''
    raise NotImplementedError


def gaussian_latitude_weights(nlat):
    '''
    Generates gaussian weights.

    Parameters
    ----------
    nlat : int
        The number of latitudes desired

    Returns
    -------
    weights : ndarray
        A one-dimensional array containing the gaussian weights.
    '''
    raise NotImplementedError


def closest_val(x, L):
    '''
    Finds the index value in an iterable closest to a desired value.

    Parameters
    ----------
    x : object
        The desired value.
    L : iterable
        The iterable in which to search for the desired value.

    Returns
    -------
    index : int
        The index of the closest value to x in L.

    Notes
    -----
    Assumes x and the entries of L are of comparable types.

    Raises
    ------
    ValueError:
        if L is empty
    '''
    # Make sure the iterable is nonempty
    if len(L) == 0:
        raise ValueError('L must not be empty')
    if isinstance(L, np.ndarray):
        # use faster numpy methods if using a numpy array
        return (np.abs(L-x)).argmin()
    # for a general iterable (like a list) we need general Python
    # start by assuming the first item is closest
    min_index = 0
    min_diff = abs(L[0] - x)
    i = 1
    while i < len(L):
        # check if each other item is closer than our current closest
        diff = abs(L[i] - x)
        if diff < min_diff:
            # if it is, set it as the new closest
            min_index = i
            min_diff = diff
        i += 1
    return min_index


def dpres_isobaric(p_lev, p_sfc, p_top, vertical_axis=None, fill_value=np.NaN):
    '''
    Calculates the pressure layer thicknesses of a constant pressure level
    coordinate system.

    Parameters
    ----------
    p_lev : ndarray
        A one dimensional array containing the constant pressure levels. May
        be in ascending or descending order.
    p_sfc : float or ndarray
        A scalar or an array containing the surface pressure data. Must have
        the same units as p_lev.
    p_top : float or ndarray
        A scalar or an array specifying the top of the column. Must have the
        same units as p_lev. If an array is given, must have the same shape
        as p_sfc.
    vertical_axis : int, optional
        The index of the returned array that should correspond to the vertical.
        Must be between 0 and the number of axes in p_sfc (inclusive).

    Returns
    -------
    dpres : ndarray
        An array specifying the pressure layer thicknesses. If p_sfc is a
        float, will be one-dimensional. Otherwise, will have one more dimension
        than p_sfc. Which axis corresponds to the vertical can be given by the
        keyword argument vertical_axis. If it is not given, the vertical axis
        will be 0 if p_sfc has 1 or 2 dimensions, or 1 if p_sfc has more
        dimensions. Note that this replicates NCL behavior for 2- and
        3-dimensional arrays. The size of the vertical dimension will be the
        same as the size of p_lev.

    See Also
    --------
    dpres_hybrid : Pressure layer thicknesses of a hybrid coordinate system

    Notes
    -----
    Calculates the layer pressure thickness of a constant pressure level
    system. At each grid point the sum of the pressure thicknesses equates to
    [p_sfc-p_top]. At each grid point, the returned values above ptop and below
    psfc will be set to fill_value. If p_top or p_sfc is between p_lev levels
    then the layer thickness is modifed accordingly.

    Raises
    ------
    ValueError
        If vertical_axis is given and is not between 0 and the number of
        axes in p_sfc.
    '''
    raise NotImplementedError


def dpres_hybrid(p_sfc, hybrid_a, hybrid_b, p0=1e5, vertical_axis=None):
    '''
    Calculates the pressure layer thicknesses of a hybrid coordinate system.

    Parameters
    ----------
    p_sfc : ndarray
        An array with surface pressure data.
    hybrid_a : ndarray
        A one-dimensional array equal to the hybrid A coefficients. Usually,
        the "interface" coefficients are input.
    hybrid_b : ndarray
        A one-dimensional array equal to the hybrid B coefficients. Usually,
        the "interface" coefficients are input.
    p0 : float, optional
        A scalar value equal to the surface reference pressure. Must have the
        same units as ps. By default, 10^5 Pa is used.
    vertical_axis : int, optional
        The index of the returned array that should correspond to the vertical.
        Must be between 0 and the number of axes in p_sfc (inclusive).

    Returns
    -------
    dpres : ndarray
        An array specifying the pressure layer thicknesses. If p_sfc is a
        float, will be one-dimensional. Otherwise, will have one more dimension
        than p_sfc. Which axis corresponds to the vertical can be given by the
        keyword argument vertical_axis. If it is not given, the vertical axis
        will be 0 if p_sfc has 1 or 2 dimensions, or 1 if p_sfc has more
        dimensions. Note that this replicates NCL behavior for 2- and
        3-dimensional arrays. The size of the vertical dimension will be the
        one less than the size of hybrid_a.

    Notes
    -----
    Calculates the layer pressure thickness of a hybrid coordinate system. At
    each grid point the sum of the pressure thicknesses equates to [psfc-ptop].
    At each grid point, the returned values above ptop and below psfc will be
    set to fill_value. If ptop or psfc is between plev levels then the layer
    thickness is modifed accordingly.

    Raises
    ------
    ValueError
        If vertical_axis is given and is not between 0 and the number of
        axes in p_sfc.
    '''
    raise NotImplementedError


def geopotential_height_hybrid(psfc, Phisfc, Tv, hyam, hybm, hyai, hybi,
                               p0=1e5, vertical_axis=None):
    '''
    Computes geopotential height in hybrid coordinates.

    Parameters
    ----------
    psfc : ndarray
        Surface pressure in Pa.
    Phisfc : ndarray
        Surface geopotential in m^2/s^2. If it is not the same shape as ps,
        then it must correspond to the rightmost dimensions of ps. May not have
        more dimensions than ps.
    Tv : ndarray
        Virtual temperature in K, ordered top-to-bottom.
    hyam: ndarray
        One-dimensional array of hybrid A coefficients (layer midpoints),
        ordered bottom-to-top.
    hybm: ndarray
        One-dimensional array of hybrid B coefficients (layer midpoints),
        ordered bottom-to-top.
    hyai: ndarray
        One-dimensional array of hybrid A coefficients (layer interfaces),
        ordered bottom-to-top.
    hybi: ndarray
        One-dimensional array of hybrid B coefficients (layer interfaces),
        ordered bottom-to-top.
    vertical_axis : int, optional
        The index of Tv that corresponds to the vertical. By default, is 0 if
        Tv has 3 or fewer axes, and 1 if Tv has more axes.

    Returns
    -------
    Phi : ndarray
        Geopotential height values. Array has the same shape as Tv.

    Notes
    -----
    Assumes no missing values in input.
    '''
    raise NotImplementedError


def hybrid_interpolate(data, ps, hybrid_a_in, hybrid_b_in, hybrid_a_out,
                       hybrid_b_out, p0=1e5, vertical_axis=None,
                       extrapolate='missing'):
    '''
    Interpolates from data on one set of hybrid levels to another set of hybrid
    levels.

    Parameters
    ----------
    data: ndarray
        Data to be interpolated.
    ps: ndarray
        Surface pressure. If given in units other than Pa, p0 should be
        specified. Its rightmost axes must correspond to the rightmost axes
        of data, not including the vertical axis of data.
    hybrid_a_in: ndarray
        Hybrid A coefficients associated with the input data.
    hybrid_b_in: ndarray
        Hybrid B coefficients associated with the input data.
    hybrid_a_out: ndarray
        Hybrid A coefficients of the returned data.
    hybrid_b_out: ndarray
        Hybrid B coefficients of the returned data.
    p0: float
        Surface reference pressure. Must be in the same units as ps.
    vertical_axis : int, optional
        The index of data that corresponds to the vertical. By default, is 0 if
        data has 3 or fewer axes, and 1 if data has more axes.
    extrapolate : str, optional
        Determines how output values outside of the range of the input axis
        should be handled. If 'missing', they are set to NaN. If 'nearest',
        they are set to the nearest input value.

    Returns
    -------
    data_out : ndarray
        data interpolated to the new hybrid vertical axis.
    '''
    raise NotImplementedError


def isobaric_to_hybrid(data, p, ps, hybrid_a, hybrid_b, p0=1e5,
                       vertical_axis=None, extrapolate='missing'):
    '''
    Interpolates data on constant pressure levels to hybrid levels.

    Parameters
    ----------
    data: ndarray
        Data to be interpolated.
    p : ndarray
        A one-dimensional array with the pressure levels of data. Must have
        the same units as ps and p0.
    ps: ndarray
        Surface pressure. If given in units other than Pa, p0 should be
        specified. Its rightmost axes must correspond to the rightmost axes
        of data, not including the vertical axis of data.
    hybrid_a: ndarray
        Hybrid A coefficients of the returned data.
    hybrid_b: ndarray
        Hybrid B coefficients of the returned data.
    p0: float
        Surface reference pressure. Must be in the same units as ps.
    vertical_axis : int, optional
        The index of data that corresponds to the vertical. By default, is 0 if
        data has 3 or fewer axes, and 1 if Tv has more axes.
    extrapolate : str, optional
        Determines how output values outside of the range of the input axis
        should be handled. If 'missing', they are set to NaN. If 'nearest',
        they are set to the nearest input value.

    Returns
    -------
    data_out : ndarray
        data interpolated to the hybrid vertical axis.
    '''
    raise NotImplementedError


def area_poly_sphere(lat, lon, r_sphere):
    '''
    Calculates the area enclosed by an arbitrary polygon on the sphere.

    Parameters
    ----------
    lat : iterable
        The latitudes, in degrees, of the vertex locations of the polygon, in
        clockwise order.
    lon : iterable
        The longitudes, in degrees, of the vertex locations of the polygon, in
        clockwise order.

    Returns
    -------
    area : float
        The desired spherical area in the same units as r_sphere.

    Notes
    -----
    This function assumes the vertices form a valid polygon (edges do not
    intersect each other).

    Reference
    ---------
    Computing the Area of a Spherical Polygon of Arbitrary Shape
    Bevis and Cambareri (1987)
    Mathematical Geology, vol.19, Issue 4, pp 335-346
    '''
    dtr = np.pi/180.

    def _tranlon(plat, plon, qlat, qlon):
        t = np.sin((qlon-plon)*dtr)*np.cos(qlat*dtr)
        b = (np.sin(qlat*dtr)*np.cos(plat*dtr) -
             np.cos(qlat*dtr)*np.sin(plat*dtr)*np.cos((qlon-plon)*dtr))
        return np.arctan2(t, b)

    if len(lat) < 3:
        raise ValueError('lat must have at least 3 vertices')
    if len(lat) != len(lon):
        raise ValueError('lat and lon must have the same length')
    total = 0.
    for i in range(-1, len(lat)-1):
        fang = _tranlon(lat[i], lon[i], lat[i+1], lon[i+1])
        bang = _tranlon(lat[i], lon[i], lat[i-1], lon[i-1])
        fvb = bang - fang
        if fvb < 0:
            fvb += 2.*np.pi
        total += fvb
    return (total - np.pi*float(len(lat)-2))*r_sphere**2


def ddx(data, axis=0, dx=None, x=None, axis_x=0, boundary='forward-backward'):
    '''
    Calculates a second-order centered finite difference derivative of data
    along the specified axis.

    Parameters
    ----------
    data : ndarray
        Data on which we are taking a derivative.
    axis : int
        Index of the data array on which to take the derivative.
    dx : float, optional
        Constant grid spacing value. Will assume constant grid spacing if
        given. May not be used with argument x. Default value is 1 unless
        x is given.
    x : ndarray, optional
        Values of the axis along which we are taking a derivative to allow
        variable grid spacing. May not be given with argument dx.
    axis_x : int, optional
        Index of the x array on which to take the derivative. Does nothing if
        x is not given as an argument.
    boundary: string, optional
        Boundary condition. If 'periodic', assume periodic boundary condition
        for centered difference. If 'forward-backward', take first-order
        forward or backward derivatives at boundary.

    Returns
    -------
    derivative : ndarray
        Derivative of the data along the specified axis.

    Raises
    ------
    ValueError:
        If an invalid boundary condition choice is given.
        If both dx and x are specified.
        If axis is out of the valid range for the shape of the data
        If x is specified and axis_x is out of the valid range for the shape
            of x
    '''
    if abs(axis) >= len(data.shape):
        raise ValueError('axis is out of bounds for the shape of data')
    if x is not None and abs(axis_x) > len(x.shape):
        raise ValueError('axis_x is out of bounds for the shape of x')
    if dx is not None and x is not None:
        raise ValueError('may not give both x and dx as keyword arguments')
    if boundary == 'periodic':
        deriv = (np.roll(data, -1, axis) - np.roll(data, 1, axis)
                 )/(np.roll(x, -1, axis_x) - np.roll(x, 1, axis_x))
    elif boundary == 'forward-backward':
        # We will take forward-backward differencing at edges
        # need some fancy indexing to handle arbitrary derivative axis
        # Initialize our index lists
        front = [slice(s) for s in data.shape]
        back = [slice(s) for s in data.shape]
        target = [slice(s) for s in data.shape]
        # Set our index values for the derivative axis
        # front is the +1 index for derivative
        front[axis] = np.array([1, -1])
        # back is the -1 index for derivative
        back[axis] = np.array([0, -2])
        # target is the position where the derivative is being calculated
        target[axis] = np.array([0, -1])
        if dx is not None:  # grid spacing is constant
            deriv = (np.roll(data, -1, axis) - np.roll(data, 1, axis))/(2.*dx)
            deriv[target] = (data[front]-data[back])/dx
        else:  # grid spacing is arbitrary
            # Need to calculate differences for our grid positions, too!
            # first take care of the centered differences
            dx = (np.roll(x, -1, axis_x) - np.roll(x, 1, axis_x))
            # now handle those annoying edge points, like with the data above
            front_x = [slice(s) for s in x.shape]
            back_x = [slice(s) for s in x.shape]
            target_x = [slice(s) for s in x.shape]
            front_x[axis_x] = np.array([1, -1])
            back_x[axis_x] = np.array([0, -2])
            target_x[axis] = np.array([0, -1])
            dx[target_x] = (x[front_x] - x[back_x])
            # Here dx spans two grid indices, no need for *2
            deriv = (np.roll(data, -1, axis) - np.roll(data, 1, axis))/dx
            deriv[target] = (data[front] - data[back])/dx
    else:  # invalid boundary condition was given
        raise ValueError('Invalid option {} for boundary '
                         'condition.'.format(boundary))
    return deriv


def d_x(data, axis, boundary='forward-backward'):
    '''
    Calculates a second-order centered finite difference of data along the
    specified axis.

    Parameters
    ----------
    data : ndarray
        Data on which we are taking a derivative.
    axis : int
        Index of the data array on which to take the difference.
    boundary: string, optional
        Boundary condition. If 'periodic', assume periodic boundary condition
        for centered difference. If 'forward-backward', take first-order
        forward or backward derivatives at boundary.

    Returns
    -------
    derivative : ndarray
        Derivative of the data along the specified axis.

    Raises
    ------
    ValueError:
        If an invalid boundary condition choice is given.
        If both dx and x are specified.
        If axis is out of the valid range for the shape of the data
        If x is specified and axis_x is out of the valid range for the shape
            of x
    '''
    if abs(axis) > len(data.shape):
        raise ValueError('axis is out of bounds for the shape of data')
    if boundary == 'periodic':
        diff = np.roll(data, -1, axis) - np.roll(data, 1, axis)
    elif boundary == 'forward-backward':
        # We will take forward-backward differencing at edges
        # need some fancy indexing to handle arbitrary derivative axis
        # Initialize our index lists
        front = [slice(s) for s in data.shape]
        back = [slice(s) for s in data.shape]
        target = [slice(s) for s in data.shape]
        # Set our index values for the derivative axis
        # front is the +1 index for derivative
        front[axis] = np.array([1, -1])
        # back is the -1 index for derivative
        back[axis] = np.array([0, -2])
        # target is the position where the derivative is being calculated
        target[axis] = np.array([0, -1])
        diff = (np.roll(data, -1, axis) - np.roll(data, 1, axis))/(2.)
        diff[target] = (data[front]-data[back])
    else:  # invalid boundary condition was given
        raise ValueError('Invalid option {} for boundary '
                         'condition.'.format(boundary))
    return diff



# -*- coding: utf-8 -*-
"""
Created on Wed Apr  1 14:29:28 2015

@author: mcgibbon
"""
import numpy as np

g0 = 9.81  # Gravitational acceleration (m/s)
stefan = 5.67e-8  # Stefan-boltzmann constant (W/m^2/K^4)
r_earth = 6370000.  # Radius of Earth (m)
Omega = 2*np.pi/86400.  # Angular velocity of Earth (Rad/s)

Rd = 287.04  # R for dry air (J/kg/K)
Rv = 461.50  # R for water vapor
Cpd = 1005.7  # Specific heat of dry air at constant pressure (J/kg/K)
Cl = 4186.  # Specific heat of liquid water (J/kg/K)
Gammad = g0/Cpd  # Dry adabatic lapse rate (K/m)

# Cpv = 1870. # Specific heat of water vapor (J/kg/K)
# Cw = 4190. # Specific heat of liquid water
Lv0 = 2.501e6  # Latent heat of vaporization for water at 0 Celsius (J/kg)
# delta = 0.608



# -*- coding: utf-8 -*-
"""
solve.py: Utilities that use equations to solve for quantities, given other
    quantities and a set of assumptions.
"""
from __future__ import division, absolute_import, unicode_literals
import inspect
from atmos import equations
import numpy as np
from six import add_metaclass
from textwrap import wrap


class ExcludeError(Exception):
    pass


def get_calculatable_quantities(inputs, methods):
    '''
    Given an interable of input quantity names and a methods dictionary,
    returns a list of output quantities that can be calculated.
    '''
    output_quantities = []
    updated = True
    while updated:
        updated = False
        for output in methods.keys():
            if output in output_quantities or output in inputs:
                # we already know we can calculate this
                continue
            for args, func in methods[output].items():
                if all([arg in inputs or arg in output_quantities
                        for arg in args]):
                    output_quantities.append(output)
                    updated = True
                    break
    return tuple(output_quantities) + tuple(inputs)


def _get_methods_that_calculate_outputs(inputs, outputs, methods):
    '''
    Given iterables of input variable names, output variable names,
    and a methods dictionary, returns the subset of the methods dictionary
    that can be calculated, doesn't calculate something we already have,
    and only contains equations that might help calculate the outputs from
    the inputs.
    '''
    # Get a list of everything that we can possibly calculate
    # This is useful in figuring out whether we can calculate arguments
    intermediates = get_calculatable_quantities(inputs, methods)
    # Initialize our return dictionary
    return_methods = {}
    # list so that we can append arguments that need to be output for
    # some of the paths
    outputs = list(outputs)
    # keep track of when to exit the while loop
    keep_going = True
    while keep_going:
        # If there are no updates in a pass, the loop will exit
        keep_going = False
        for output in outputs:
            try:
                output_dict = return_methods[output]
            except:
                output_dict = {}
            for args, func in methods[output].items():
                # only check the method if we're not already returning it
                if args not in output_dict.keys():
                    # Initialize a list of intermediates needed to use
                    # this method, to add to outputs if we find we can
                    # use it.
                    needed = []
                    for arg in args:
                        if arg in inputs:
                            # we have this argument
                            pass
                        elif arg in outputs:
                            # we may need to calculate one output using
                            # another output
                            pass
                        elif arg in intermediates:
                            if arg not in outputs:
                                # don't need to add to needed if it's already
                                # been put in outputs
                                needed.append(arg)
                        else:
                            # Can't do this func
                            break
                    else:  # did not break, can calculate this
                        output_dict[args] = func
                        if len(needed) > 0:
                            # We added an output, so need another loop
                            outputs.extend(needed)
                            keep_going = True
            if len(output_dict) > 0:
                return_methods[output] = output_dict
    return return_methods


def _get_calculatable_methods_dict(inputs, methods):
    '''
    Given an iterable of input variable names and a methods dictionary,
    returns the subset of that methods dictionary that can be calculated and
    which doesn't calculate something we already have. Additionally it may
    only contain one method for any given output variable, which is the one
    with the fewest possible arguments.
    '''
    # Initialize a return dictionary
    calculatable_methods = {}
    # Iterate through each potential method output
    for var in methods.keys():
        # See if we already have this output
        if var in inputs:
            continue  # if we have it, we don't need to calculate it!
        else:
            # Initialize a dict for this output variable
            var_dict = {}
            for args, func in methods[var].items():
                # See if we have what we need to solve this equation
                if all([arg in inputs for arg in args]):
                    # If we do, add it to the var_dict
                    var_dict[args] = func
            if len(var_dict) == 0:
                # No methods for this variable, keep going
                continue
            elif len(var_dict) == 1:
                # Exactly one method, perfect.
                calculatable_methods[var] = var_dict
            else:
                # More than one method, find the one with the least arguments
                min_args = min(var_dict.keys(), key=lambda x: len(x))
                calculatable_methods[var] = {min_args: var_dict[min_args]}
    return calculatable_methods


def _get_shortest_solution(outputs, inputs, exclude, methods):
    '''
    Parameters
    ----------
    outputs: tuple
        Strings corresponding to the final variables we want output
    inputs: tuple
        Strings corresponding to the variables we have so far
    exclude: tuple
        Strings corresponding to variables that won't help calculate the
        outputs.
    methods: dict
        A methods dictionary

    Returns (funcs, func_args, extra_values).
    '''
    methods = _get_methods_that_calculate_outputs(inputs, outputs,
                                                  methods)
    calculatable_methods = _get_calculatable_methods_dict(inputs, methods)
    # Check if we can already directly compute the outputs
    if all([(o in calculatable_methods.keys()) or (o in inputs)
            for o in outputs]):
        funcs = []
        args = []
        extra_values = []
        for o in outputs:
            if o not in inputs:
                o_args, o_func = list(calculatable_methods[o].items())[0]
                funcs.append(o_func)
                args.append(o_args)
                extra_values.append(o)
        return tuple(funcs), tuple(args), tuple(extra_values)
    # Check if there's nothing left to try to calculate
    if len(calculatable_methods) == 0:
        raise ValueError('cannot calculate outputs from inputs')
    next_variables = [key for key in calculatable_methods.keys()
                      if key not in exclude]
    if len(next_variables) == 0:
        raise ExcludeError
    results = []
    intermediates = []
    for i in range(len(next_variables)):
        try:
            results.append(_get_shortest_solution(
                outputs, inputs + (next_variables[i],), exclude, methods))
        except ExcludeError:
            continue
        intermediates.append(next_variables[i])
        exclude = exclude + (next_variables[i],)
    if len(results) == 0:
        # all subresults raised ExcludeError
        raise ExcludeError

    def option_key(a):
        return len(a[0]) + 0.001*len(a[1])
    best_result = min(results, key=option_key)
    best_index = results.index(best_result)
    best_intermediate = intermediates[best_index]
    args, func = list(calculatable_methods[best_intermediate].items())[0]
    best_option = ((func,) + best_result[0], (args,) + best_result[1],
                   (best_intermediate,) + best_result[2])
    return best_option


def _get_module_methods(module):
    '''
    Returns a methods list corresponding to the equations in the given
    module. Each entry is a dictionary with keys 'output', 'args', and
    'func' corresponding to the output, arguments, and function of the
    method. The entries may optionally include 'assumptions' and
    'overridden_by_assumptions' as keys, stating which assumptions are
    required to use the method, and which assumptions mean the method
    should not be used because it is overridden.
    '''
    # Set up the methods dict we will eventually return
    methods = []
    funcs = []
    for item in inspect.getmembers(equations):
        if (item[0][0] != '_' and '_from_' in item[0]):
            func = item[1]
            output = item[0][:item[0].find('_from_')]
        # avoid returning duplicates
        if func in funcs:
            continue
        else:
            funcs.append(func)
        args = tuple(inspect.getargspec(func).args)
        try:
            assumptions = tuple(func.assumptions)
        except AttributeError:
            raise NotImplementedError('function {0} in equations module has no'
                                      ' assumption '
                                      'definition'.format(func.__name__))
        try:
            overridden_by_assumptions = func.overridden_by_assumptions
        except AttributeError:
            overridden_by_assumptions = ()
        methods.append({
            'func': func,
            'args': args,
            'output': output,
            'assumptions': assumptions,
            'overridden_by_assumptions': overridden_by_assumptions,
        })
    return methods


def _fill_doc(s, module, default_assumptions):
        assumptions = module.assumptions
        s = s.replace(
            '<assumptions list goes here>',
            '\n'.join(sorted(
                ["'{0}': {1}".format(a, desc) for a, desc in
                 assumptions.items()],
                key=lambda x: x.lower())))
        s = s.replace(
            '<default assumptions list goes here>',
            '\n'.join(
                wrap('Default assumptions are ' +
                     ', '.join(["'{0}'".format(a) for a in
                                default_assumptions]) + '.', width=80)))
        s = s.replace(
            '<quantity parameter list goes here>',
            '\n'.join(sorted([
                '{0} {1} ({2})'.format(
                    (q + ' :').ljust(9), info['name'], info['units'])
                for q, info in
                module.quantities.items()
                ], key=lambda x: x.lower())))
        return s


def _check_scalar(value):
    '''If value is a 0-dimensional array, returns the contents of value.
       Otherwise, returns value.
    '''
    if isinstance(value, np.ndarray):
        if value.ndim == 0:
            # We have a 0-dimensional array
            return value[None][0]
    return value


# We need to define a MetaClass in order to have dynamic docstrings for our
# Solver objects, generated based on the equations module
class SolverMeta(type):
    '''
Metaclass for BaseSolver to automatically generate docstrings and assumption
lists for subclasses of BaseSolver.
    '''

    def __new__(cls, name, parents, dct):
        if dct['_equation_module'] is not None:
            # Update the class docstring
            if '__doc__' in dct.keys():
                dct['__doc__'] = _fill_doc(
                    dct['__doc__'], dct['_equation_module'],
                    dct['default_assumptions'])

            assumptions = set([])
            for f in inspect.getmembers(equations):
                try:
                    assumptions.update(f[1].assumptions)
                except AttributeError:
                    pass
            dct['all_assumptions'] = tuple(assumptions)

        # we need to call type.__new__ to complete the initialization
        instance = super(SolverMeta, cls).__new__(cls, name, parents, dct)
        return instance


@add_metaclass(SolverMeta)
class BaseSolver(object):
    '''
Base class for solving systems of equations. Should not be instantiated,
as it is not associated with any equations.
    '''

    _equation_module = None

    def __init__(self, **kwargs):
        if self._equation_module is None:
            raise NotImplementedError('Class needs _equation_module '
                                      'defined')
        if 'debug' in kwargs.keys():
            self._debug = kwargs.pop('debug')
        else:
            self._debug = False
        # See if an assumption set was given
        if 'assumptions' in kwargs.keys():
            # If it was, make sure it wasn't given with other ways of
            # setting assumptions (by modifying the default assumptions)
            if ('add_assumptions' in kwargs.keys() or
                    'remove_assumptions' in kwargs.keys()):
                raise ValueError('cannot give kwarg assumptions with '
                                 'add_assumptions or remove_assumptions')
            assumptions = kwargs.pop('assumptions')
        else:
            # if it wasn't, modify the default assumptions
            assumptions = self.default_assumptions
            if 'add_assumptions' in kwargs.keys():
                if 'remove_assumptions' in kwargs.keys():
                    # make sure there is no overlap
                    if any([a in kwargs['remove_assumptions']
                            for a in kwargs['add_assumptions']]):
                        raise ValueError('assumption may not be present in '
                                         'both add_assumptions and '
                                         'remove_assumptions')
                # add assumptions, avoiding duplicates
                assumptions = assumptions + tuple(
                    [a for a in kwargs.pop('add_assumptions') if a not in
                     assumptions])
            if 'remove_assumptions' in kwargs.keys():
                # remove assumptions if present
                remove_assumptions = kwargs.pop('remove_assumptions')
                self._ensure_assumptions(*assumptions)
                assumptions = tuple([a for a in assumptions if a not in
                                     remove_assumptions])
        # Make sure all set assumptions are valid (not misspelt, for instance)
        self._ensure_assumptions(*assumptions)
        # now that we have our assumptions, use them to set the methods
        self.methods = self._get_methods(assumptions)
        # make sure the remaining variables are quantities
        self._ensure_quantities(*kwargs.keys())
        # also store the quantities
        self.vars = kwargs

    def _ensure_assumptions(self, *args):
        '''Raises ValueError if any of the args are not strings corresponding
           to short forms of assumptions for this Solver.
        '''
        for arg in args:
            if arg not in self.all_assumptions:
                raise ValueError('{0} does not correspond to a valid '
                                 'assumption.'.format(arg))

    def _ensure_quantities(self, *args):
        '''Raises ValueError if any of the args are not strings corresponding
           to quantity abbreviations for this Solver.
        '''
        for arg in args:
            if arg not in self._equation_module.quantities.keys():
                raise ValueError('{0} does not correspond to a valid '
                                 'quantity.'.format(arg))

    def calculate(self, *args):
        '''
Calculates and returns a requested quantity from quantities stored in this
object at initialization.

Parameters
----------
varname_out : string
    Name of quantity to be calculated.

Returns
-------
quantity : ndarray
    Calculated quantity, in units listed under quantity parameters.

Notes
-----
See the documentation for this object for a complete list of quantities
that may be calculated, in the "Quantity Parameters" section.

Raises
------
ValueError:
    If the output quantity cannot be determined from the input
    quantities.
        '''
        self._ensure_quantities(*args)
        possible_quantities = get_calculatable_quantities(self.vars.keys(),
                                                          self.methods)
        for arg in args:
            if arg not in possible_quantities:
                raise ValueError('cannot calculate {0} from inputs'.format(
                    arg))
        funcs, func_args, extra_values = \
            _get_shortest_solution(tuple(args), tuple(self.vars.keys()), (),
                                   self.methods)
        # Above method completed successfully if no ValueError has been raised
        # Calculate each quantity we need to calculate in order
        for i, func in enumerate(funcs):
            # Compute this quantity
            value = func(*[self.vars[varname] for varname in func_args[i]])
            # Add it to our dictionary of quantities for successive functions
            self.vars[extra_values[i]] = value
        if self._debug:
            # We should return a list of funcs as the last item returned
            if len(args) == 1:
                return _check_scalar(self.vars[args[0]]), funcs
            else:
                return ([_check_scalar(self.vars[arg]) for arg in args] +
                        [funcs, ])
        else:
            # no function debugging, just return the quantities
            if len(args) == 1:
                return _check_scalar(self.vars[args[0]])
            else:
                return [_check_scalar(self.vars[arg]) for arg in args]

    def _get_methods(self, assumptions):
        '''
Returns a dictionary of methods including the default methods of the
class and specified optional methods. Will override a default method
if an optional method is given that takes the same inputs and produces
the same output.

Parameters
----------
methods: iterable
    Strings referring to optional methods in self.optional_methods.

Returns
-------
methods : dict
    A dictionary whose keys are strings indicating output variables,
    and values are dictionaries indicating equations for that output. The
    equation dictionary's keys are strings indicating variables to use
    as function arguments, and its values are the functions themselves.

Raises
------
ValueError
    If a method given is not present in self.optional_methods.
    If multiple optional methods are selected which calculate the same
    output quantity from the same input quantities.
        '''
        # make sure all assumptions actually apply to equations
        # this will warn the user of typos
        for a in assumptions:
            if a not in self.all_assumptions:
                raise ValueError('assumption {0} matches no '
                                 'equations'.format(a))
        # create a dictionary to which we will add methods
        methods = {}
        # get a set of all the methods in the module
        module_methods = _get_module_methods(self._equation_module)
        # Go through each output variable
        for dct in module_methods:
            # Make sure this method is not overridden
            if any(item in assumptions for item in
                    dct['overridden_by_assumptions']):
                continue
            # Make sure all assumptions of the method are satisfied
            elif all(item in assumptions for item in dct['assumptions']):
                # Make sure we have a dict entry for this output quantity
                if dct['output'] not in methods.keys():
                    methods[dct['output']] = {}
                # Make sure we aren't defining methods with same signature
                if dct['args'] in methods[dct['output']].keys():
                    raise ValueError(
                        'assumptions given define duplicate '
                        'equations {0} and {1}'.format(
                            str(dct['func']),
                            str(methods[dct['output']][dct['args']])))
                # Add the method to the methods dict
                methods[dct['output']][dct['args']] = dct['func']
        return methods


class FluidSolver(BaseSolver):
    '''
Initializes with the given assumptions enabled, and variables passed as
keyword arguments stored.

Parameters
----------
assumptions : tuple, optional
    Strings specifying which assumptions to enable. Overrides the default
    assumptions. See below for a list of default assumptions.
add_assumptions : tuple, optional
    Strings specifying assumptions to use in addition to the default
    assumptions. May not be given in combination with the assumptions kwarg.
remove_assumptions : tuple, optional
    Strings specifying assumptions not to use from the default assumptions.
    May not be given in combination with the assumptions kwarg. May not
    contain strings that are contained in add_assumptions, if given.
quantity : ndarray, optional
    Keyword arguments used to pass in arrays of data that correspond to
    quantities used for calculations. For a complete list of kwargs that
    may be used, see the Quantity Parameters section below.

Quantity Parameters
-------------------
<quantity parameter list goes here>

Returns
-------
out : FluidSolver
    A FluidSolver object with the specified assumptions and variables.

Assumptions
-----------
<default assumptions list goes here>

Assumption descriptions:
<assumptions list goes here>

Examples
--------
>>> solver = FluidSolver(rho=array1, p=array2)

Non-default assumptions:

>>> solver = FluidSolver(add_assumptions=('Tv equals T'), rho=array1, p=array2)
    '''

    # module containing fluid dynamics equations
    _equation_module = equations
    # which assumptions to use by default
    default_assumptions = (
        'ideal gas', 'hydrostatic', 'constant g', 'constant Lv', 'constant Cp',
        'no liquid water', 'no ice', 'bolton',)


def calculate(*args, **kwargs):
    '''
Calculates and returns a requested quantity from quantities passed in as
keyword arguments.

Parameters
----------
args : string
    Names of quantities to be calculated.
assumptions : tuple, optional
    Strings specifying which assumptions to enable. Overrides the default
    assumptions. See below for a list of default assumptions.
add_assumptions : tuple, optional
    Strings specifying assumptions to use in addition to the default
    assumptions. May not be given in combination with the assumptions kwarg.
remove_assumptions : tuple, optional
    Strings specifying assumptions not to use from the default assumptions.
    May not be given in combination with the assumptions kwarg. May not
    contain strings that are contained in add_assumptions, if given.
quantity : ndarray, optional
    Keyword arguments used to pass in arrays of data that correspond to
    quantities used for calculations. For a complete list of kwargs that
    may be used, see the Quantity Parameters section below.

Assumptions
-----------
<default assumptions list goes here>

Assumption descriptions:
<assumptions list goes here>

Quantity Parameters
-------------------
<quantity parameter list goes here>

Returns
-------
quantity : ndarray or iris Cube
    Calculated quantity.
    Return type is the same as quantity parameter types.
    If multiple quantities are requested, returns a tuple containing the
    quantities.

Notes
-----
Calculating multiple quantities at once can avoid re-computing intermediate
quantities, but requires more memory.

Examples
--------
>>>
    '''
    if len(args) == 0:
        raise ValueError('must specify quantities to calculate')
    # initialize a solver to do the work
    solver = FluidSolver(**kwargs)
    # get the output
    return solver.calculate(*args)

# autocomplete some sections of the docstring for calculate
calculate.__doc__ = _fill_doc(calculate.__doc__, equations,
                              FluidSolver.default_assumptions)



from setuptools import setup

setup(
    name='atmos',
    packages=['atmos'],
    version='0.2.1',
    description='Atmospheric sciences utility library',
    author='Jeremy McGibbon',
    author_email='mcgibbon@uw.edu',
    keywords=['atmos', 'atmospheric', 'equations', 'geoscience', 'science'],
    classifiers=[],
)



# -*- coding: utf-8 -*-
"""
test.py: Testing suite for other modules.
"""
from __future__ import division, unicode_literals
import unittest
import nose
import numpy as np
import inspect
from atmos import equations
from atmos import util
from atmos import decorators
from nose.tools import raises
from atmos.constants import Rd
from atmos.solve import BaseSolver, FluidSolver, calculate, \
    _get_module_methods, _get_calculatable_methods_dict,\
    _get_shortest_solution
from atmos.util import quantity_string, assumption_list_string, \
    quantity_spec_string, doc_paragraph, \
    strings_to_list_string


def test_quantities_dict_complete():
    names = [item['output'] for item in _get_module_methods(equations)]
    for name in names:
        if name not in equations.quantities.keys():
            try:
                util.parse_derivative_string(name)
            except ValueError:
                raise AssertionError('{} not in quantities dict'.format(name))


def test_get_module_methods_nonempty():
    result = _get_module_methods(equations)
    assert len(result) > 0


def test_default_assumptions_exist():
    for m in FluidSolver.default_assumptions:
        if m not in FluidSolver.all_assumptions:
            raise AssertionError('{} not a valid assumption'.format(m))


class ddxTests(unittest.TestCase):

    def setUp(self):
        self.data = np.zeros((2, 3))
        self.data[:] = np.array([0., 5., 10.])[None, :]
        self.axis1 = np.array([5., 6., 7.])
        self.axis2 = np.array([[5., 6., 7.], [8., 10., 12.]])
        self.deriv1 = 5.*np.ones((2, 3))
        self.deriv2 = 5.*np.ones((2, 3))
        self.deriv2[1, :] *= 0.5

    @raises(ValueError)
    def test_invalid_data_axis(self):
        util.ddx(self.data, axis=2)

    @raises(ValueError)
    def test_invalid_axis_x(self):
        util.ddx(self.data, axis=2, x=self.axis1, axis_x=1)


class ClosestValTests(unittest.TestCase):

    def setUp(self):
        self.list = [1., 5., 10.]
        self.array = np.array([1., 5., 10.])

    def tearDown(self):
        self.list = None
        self.array = None

    @raises(ValueError)
    def testValueErrorOnEmptyList(self):
        util.closest_val(1., [])

    @raises(ValueError)
    def testValueErrorOnEmptyArray(self):
        util.closest_val(1., np.array([]))

    def testClosestValSingle(self):
        val = util.closest_val(1., np.array([50.]))
        assert val == 0

    def testClosestValInternal(self):
        val = util.closest_val(3.5, self.array)
        assert val == 1

    def testClosestValBelow(self):
        val = util.closest_val(-5., self.array)
        assert val == 0

    def testClosestValAbove(self):
        val = util.closest_val(20., self.array)
        assert val == 2

    def testClosestValInternalNegative(self):
        val = util.closest_val(-3.5, -1*self.array)
        assert val == 1

    def testClosestValInternalList(self):
        val = util.closest_val(3.5, self.list)
        assert val == 1

    def testClosestValBelowList(self):
        val = util.closest_val(-5., self.list)
        assert val == 0

    def testClosestValAboveList(self):
        val = util.closest_val(20., self.list)
        assert val == 2


class AreaPolySphereTests(unittest.TestCase):

    def setUp(self):
        self.lat1 = np.array([0.0, 90.0, 0.0])
        self.lon1 = np.array([0.0, 0.0, 90.0])
        self.area1 = 1.5708
        self.tol1 = 0.0001

    @raises(ValueError)
    def test_area_poly_sphere_insufficient_vertices(self):
        util.area_poly_sphere(self.lat1[:2], self.lon1[:2], 1.)

    def test_area_poly_sphere(self):
        area_calc = util.area_poly_sphere(self.lat1, self.lon1, 1.)
        assert abs(area_calc - self.area1) < self.tol1

    def test_area_poly_sphere_different_radius(self):
        area_calc = util.area_poly_sphere(self.lat1, self.lon1, 2.)
        assert abs(area_calc - 4*self.area1) < 4*self.tol1


class DecoratorTests(unittest.TestCase):

    def setUp(self):
        def dummyfunction(x):
            '''Dummy docstring'''
            return x
        self.func = dummyfunction
        self.func_name = dummyfunction.__name__
        self.func_argspec = inspect.getargspec(dummyfunction)
        self.quantity_dict = {
            'T': {'name': 'air temperature', 'units': 'K'},
            'qv': {'name': 'specific humidity', 'units': 'kg/kg'},
            'p': {'name': 'air pressure', 'units': 'Pa'},
        }
        self.assumption_dict = {
            'a1': 'a1_long',
            'a2': 'a2_long',
            'a3': 'a3_long',
        }
        self.func_argspec = inspect.getargspec(self.func)

    def tearDown(self):
        self.func = None
        self.quantity_dict = None
        self.assumption_dict = None

    def test_assumes_empty(self, **kwargs):
        func = decorators.assumes()(self.func)
        assert func.assumptions == ()
        assert func.__name__ == self.func_name
        assert inspect.getargspec(func) == self.func_argspec

    def test_assumes_single(self, **kwargs):
        func = decorators.assumes('a1')(self.func)
        assert func.assumptions == ('a1',)
        assert func.__name__ == self.func_name
        assert inspect.getargspec(func) == self.func_argspec

    def test_assumes_triple(self, **kwargs):
        func = decorators.assumes('a1', 'a2', 'a3')(self.func)
        assert func.assumptions == ('a1', 'a2', 'a3',)
        assert func.__name__ == self.func_name
        assert inspect.getargspec(func) == self.func_argspec

    def test_overridden_by_assumptions_empty(self, **kwargs):
        func = decorators.overridden_by_assumptions()(self.func)
        assert func.overridden_by_assumptions == ()
        assert func.__name__ == self.func_name
        assert inspect.getargspec(func) == self.func_argspec

    def test_overridden_by_assumptions_single(self, **kwargs):
        func = decorators.overridden_by_assumptions('a1')(self.func)
        assert func.overridden_by_assumptions == ('a1',)
        assert func.__name__ == self.func_name
        assert inspect.getargspec(func) == self.func_argspec

    def test_overridden_by_assumptions_triple(self, **kwargs):
        func = decorators.overridden_by_assumptions(
            'a1', 'a2', 'a3')(self.func)
        assert func.overridden_by_assumptions == ('a1', 'a2', 'a3',)
        assert func.__name__ == self.func_name
        assert inspect.getargspec(func) == self.func_argspec

    @raises(ValueError)
    def test_autodoc_invalid_no_extra_args(self):
        def invalid_function(T, p):
            return T
        decorators.equation_docstring(
            self.quantity_dict, self.assumption_dict)(invalid_function)

    @raises(ValueError)
    def test_autodoc_invalid_args_function(self):
        def T_from_x_y(x, y):
            return x
        decorators.equation_docstring(
            self.quantity_dict, self.assumption_dict)(T_from_x_y)

    @raises(ValueError)
    def test_autodoc_invalid_extra_args(self):
        def invalid_function(T, p):
            return T
        decorators.equation_docstring(
            self.quantity_dict, self.assumption_dict, equation='x=y',
            references='reference here', notes='c sharp')(invalid_function)

    def test_autodoc_valid(self):
        def T_from_p_qv(p, qv):
            return 1.
        func = decorators.equation_docstring(
            self.quantity_dict, self.assumption_dict, equation='x=y',
            references='ref', notes='notes')(T_from_p_qv)
        assert func.__doc__ == (
            'Calculates air temperature (K).\n\nx=y\n\nPa'
            'rameters\n----------\np : float or ndarray\n    Data for air pre'
            'ssure (Pa).\nqv : float or ndarray\n    Data for specific humidi'
            'ty (kg/kg).\n\nReturns\n-------\nT : float or ndarray\n    Data '
            'for air temperature (K).\n\nNotes\n-----\nnotes\n\nReferences\n-'
            '---------\nref\n')

    def test_autodoc_valid_assumes(self):
        @decorators.assumes('a1')
        def T_from_p_qv(p, qv):
            return 1.
        func = decorators.equation_docstring(
            self.quantity_dict, self.assumption_dict, equation='x=y',
            references='ref', notes='notes')(T_from_p_qv)
        assert func.__doc__ == (
            'Calculates air temperature (K) assuming a1_long.\n\nx=y\n\nPa'
            'rameters\n----------\np : float or ndarray\n    Data for air pre'
            'ssure (Pa).\nqv : float or ndarray\n    Data for specific humidi'
            'ty (kg/kg).\n\nReturns\n-------\nT : float or ndarray\n    Data '
            'for air temperature (K).\n\nNotes\n-----\nnotes\n\nReferences\n-'
            '---------\nref\n')


class StringUtilityTests(unittest.TestCase):

    def setUp(self):
        self.quantity_dict = {
            'T': {'name': 'air temperature', 'units': 'K'},
            'qv': {'name': 'specific humidity', 'units': 'kg/kg'},
            'p': {'name': 'air pressure', 'units': 'Pa'},
        }
        self.assumption_dict = {
            'a1': 'a1_long',
            'a2': 'a2_long',
            'a3': 'a3_long',
        }

    def tearDown(self):
        self.quantity_dict = None
        self.assumption_dict = None

    def test_quantity_string(self):
        string = quantity_string('T', self.quantity_dict)
        assert string == 'air temperature (K)'

    @raises(ValueError)
    def test_quantity_string_invalid_quantity(self):
        quantity_string('rhombus', self.quantity_dict)

    @raises(ValueError)
    def test_strings_to_list_string_empty(self):
        string = strings_to_list_string(())
        assert string == ''

    @raises(TypeError)
    def test_strings_to_list_string_input_string(self):
        strings_to_list_string('hello')

    def test_strings_to_list_string_single(self):
        string = strings_to_list_string(('string1',))
        assert string == 'string1'

    def test_strings_to_list_string_double(self):
        string = strings_to_list_string(('string1', 'string2'))
        assert string == 'string1 and string2'

    def test_strings_to_list_string_triple(self):
        string = strings_to_list_string(('string1', 'string2', 'string3'))
        assert string == 'string1, string2, and string3'

    @raises(ValueError)
    def test_assumption_list_string_empty(self):
        assumption_list_string((), self.assumption_dict)

    @raises(TypeError)
    def test_assumption_list_string_input_string(self):
        assumption_list_string('hello', self.assumption_dict)

    @raises(ValueError)
    def test_assumption_list_string_invalid(self):
        assumption_list_string(('ooglymoogly',), self.assumption_dict)

    def test_assumption_list_string_single(self):
        string = assumption_list_string(('a1',), self.assumption_dict)
        assert string == 'a1_long'

    def test_assumption_list_string_double(self):
        string = assumption_list_string(('a1', 'a2'), self.assumption_dict)
        assert string == 'a1_long and a2_long'

    def test_assumption_list_string_triple(self):
        string = assumption_list_string(('a1', 'a2', 'a3'),
                                        self.assumption_dict)
        assert string == 'a1_long, a2_long, and a3_long'

    @raises(ValueError)
    def test_quantity_spec_string_invalid(self):
        quantity_spec_string('rv', self.quantity_dict)

    def test_quantity_spec_string_valid(self):
        string = quantity_spec_string('T', self.quantity_dict)
        assert string == ('T : float or ndarray\n'
                          '    Data for air temperature (K).')

    def test_doc_paragraph_no_wrap_for_short_string(self):
        string = doc_paragraph('The quick brown fox jumped over the yellow '
                               'doge. The quick brown fox jumped over')
        assert '\n' not in string

    def test_doc_paragraph_wraps_on_word(self):
        string = doc_paragraph('The quick brown fox jumped over the yellow '
                               'doge. The quick brown fox jumped over.')
        if (string != 'The quick brown fox jumped over the yellow doge. '
                'The quick brown fox jumped\nover.'):
            raise AssertionError('incorrect string "{0}"'.format(string))

    def test_doc_paragraph_indent_wrap(self):
        string = doc_paragraph('The quick brown fox jumped over the yellow '
                               'doge. The quick brown fox jumped over',
                               indent=1)
        if (string != ' The quick brown fox jumped over the yellow doge. '
                'The quick brown fox jumped\n over'):
            raise AssertionError('incorrect string "{0}"'.format(string))

    def test_doc_paragraph_zero_indent(self):
        string = doc_paragraph('The quick brown fox jumped over the yellow '
                               'doge. The quick brown fox jumped over.',
                               indent=0)
        if (string != 'The quick brown fox jumped over the yellow doge. '
                'The quick brown fox jumped\nover.'):
            raise AssertionError('incorrect string "{0}"'.format(string))


class ParseDerivativeStringTests(unittest.TestCase):

    def setUp(self):
        self.quantity_dict = {
            'T': {'name': 'air temperature', 'units': 'K'},
            'qv': {'name': 'specific humidity', 'units': 'kg/kg'},
            'p': {'name': 'air pressure', 'units': 'Pa'},
        }

    def tearDown(self):
        self.quantity_dict = None

    @raises(ValueError)
    def test_invalid_format(self):
        util.parse_derivative_string('ooglymoogly', self.quantity_dict)

    @raises(ValueError)
    def test_invalid_variable(self):
        util.parse_derivative_string('dpdz', self.quantity_dict)

    def test_dTdp(self):
        var1, var2 = util.parse_derivative_string('dTdp', self.quantity_dict)
        assert var1 == 'T'
        assert var2 == 'p'

    @raises(ValueError)
    def test_dpdT(self):
        util.parse_derivative_string('dpdT', self.quantity_dict)

class OverriddenByAssumptionsTests(unittest.TestCase):
    def test_overridden_by_assumptions_empty(self):
        @decorators.overridden_by_assumptions()
        def foo():
            return None
        assert foo.overridden_by_assumptions == ()

    def test_overridden_by_assumptions_single(self):
        @decorators.overridden_by_assumptions('test assumption')
        def foo():
            return None
        assert foo.overridden_by_assumptions == ('test assumption',)

    def test_overridden_by_assumptions_multiple(self):
        @decorators.overridden_by_assumptions('test assumption', 'a2')
        def foo():
            return None
        assert foo.overridden_by_assumptions == ('test assumption', 'a2')


class GetCalculatableMethodsDictTests(unittest.TestCase):

    def test_get_calculatable_methods_dict_empty(self):
        methods = {}
        out_methods = _get_calculatable_methods_dict((), methods)
        assert isinstance(out_methods, dict)
        assert len(out_methods) == 0

    def test_get_calculatable_methods_dict_returns_correct_type(self):
        methods = {'a': {('b',): lambda x: x}}
        out_methods = _get_calculatable_methods_dict(('b',), methods)
        assert isinstance(out_methods, dict)

    def test_get_calculatable_methods_dict_gets_single_method(self):
        methods = {'a': {('b',): lambda x: x}}
        out_methods = _get_calculatable_methods_dict(('b',), methods)
        assert 'a' in out_methods.keys()

    def test_get_calculatable_methods_dict_removes_correct_second_method(
            self):
        methods = {'a': {('b',): lambda x: x,
                         ('c', 'b'): lambda x: x}}
        out_methods = _get_calculatable_methods_dict(('b', 'c'), methods)
        assert 'a' in out_methods.keys()
        assert ('b',) in out_methods['a'].keys()
        assert len(out_methods['a']) == 1

    def test_get_calculatable_methods_dict_removes_irrelevant_second_method(
            self):
        methods = {'a': {('b', 'd'): lambda x, y: x,
                         ('c',): lambda x: x}}
        out_methods = _get_calculatable_methods_dict(('b', 'd'), methods)
        assert 'a' in out_methods.keys()
        assert ('b', 'd') in out_methods['a'].keys()
        assert len(out_methods['a']) == 1

    def test_get_calculatable_methods_dict_gets_no_methods(self):
        methods = {'a': {('b',): lambda x: x},
                   'x': {('y', 'z'): lambda y, z: y*z}
                   }
        out_methods = _get_calculatable_methods_dict(('q',), methods)
        assert isinstance(out_methods, dict)
        assert len(out_methods) == 0

    def test_get_calculatable_methods_dict_doesnt_calculate_input(self):
        methods = {'a': {('b',): lambda x: x},
                   'x': {('y', 'z'): lambda y, z: y*z}
                   }
        out_methods = _get_calculatable_methods_dict(('b', 'a'), methods)
        assert isinstance(out_methods, dict)
        assert len(out_methods) == 0


class BaseSolverTests(unittest.TestCase):

    @raises(NotImplementedError)
    def test_cannot_instantiate(self):
        BaseSolver()


class FluidSolverTests(unittest.TestCase):

    def setUp(self):
        shape = (3, 4, 2, 2)
        self.vars1 = {'Tv': np.ones(shape),
                      'p': np.ones(shape)}
        self.vars2 = {'T': np.ones(shape),
                      'p': np.ones(shape)}

    def tearDown(self):
        self.vars1 = None
        self.vars2 = None

    def test_creation_no_arguments(self):
        FluidSolver()

    def test_is_instance_of_BaseSolver(self):
        deriv = FluidSolver()
        assert isinstance(deriv, BaseSolver)

    def test_creation_one_method(self):
        FluidSolver(assumptions=('hydrostatic',))

    def test_creation_compatible_methods(self):
        FluidSolver(assumptions=('hydrostatic', 'Tv equals T',))

    @raises(ValueError)
    def test_creation_incompatible_methods(self):
        FluidSolver(assumptions=('Goff-Gratch', 'Wexler',))

    @raises(ValueError)
    def test_creation_undefined_method(self):
        FluidSolver(assumptions=('moocow',))

    @raises(ValueError)
    def test_creation_undefined_method_with_defined_method(self):
        FluidSolver(assumptions=('hydrostatic', 'moocow',))

    def test_creation_with_vars(self):
        FluidSolver(**self.vars1)

    def test_creation_with_vars_and_method(self):
        FluidSolver(assumptions=('Tv equals T',), **self.vars1)

    def test_simple_calculation(self):
        deriver = FluidSolver(**self.vars1)
        rho = deriver.calculate('rho')
        assert (rho == 1/Rd).all()
        assert isinstance(rho, np.ndarray)

    def test_depth_2_calculation(self):
        deriver = FluidSolver(assumptions=FluidSolver.default_assumptions +
                              ('Tv equals T',), **self.vars2)
        rho = deriver.calculate('rho')
        assert (rho == 1/Rd).all()
        assert isinstance(rho, np.ndarray)

    def test_double_calculation(self):
        deriver = FluidSolver(add_assumptions=('Tv equals T',), **self.vars2)
        Tv = deriver.calculate('Tv')
        rho = deriver.calculate('rho')
        assert (rho == 1/Rd).all()
        assert isinstance(rho, np.ndarray)
        assert isinstance(Tv, np.ndarray)

    def test_double_reverse_calculation(self):
        deriver = FluidSolver(add_assumptions=('Tv equals T',), **self.vars2)
        rho = deriver.calculate('rho')
        print('now Tv')
        Tv = deriver.calculate('Tv')
        assert (rho == 1/Rd).all()
        assert isinstance(rho, np.ndarray)
        assert isinstance(Tv, np.ndarray)


class calculateTests(unittest.TestCase):

    def setUp(self):
        self.shape = (3, 4, 2, 2)
        self.vars1 = {'Tv': np.ones(self.shape),
                      'p': np.ones(self.shape)}
        self.vars2 = {'T': np.ones(self.shape),
                      'p': np.ones(self.shape)}

    def tearDown(self):
        self.vars1 = None
        self.vars2 = None

    def test_simple_calculation(self):
        rho = calculate('rho', **self.vars1)
        assert (rho.shape == self.shape)
        assert (rho == 1/Rd).all()
        assert isinstance(rho, np.ndarray)

    def test_returns_float(self):
        rho = calculate('rho', Tv=1., p=1.)
        assert isinstance(rho, float)

    def test_depth_2_calculation(self):
        rho = calculate('rho', add_assumptions=('Tv equals T',), **self.vars2)
        assert rho.shape == self.shape
        assert (rho == 1/Rd).all()
        assert isinstance(rho, np.ndarray)

    def test_double_calculation(self):
        Tv, rho = calculate('Tv', 'rho', add_assumptions=('Tv equals T',),
                            **self.vars2)
        assert Tv.shape == self.shape
        assert rho.shape == self.shape
        assert (rho == 1/Rd).all()
        assert isinstance(rho, np.ndarray)
        assert isinstance(Tv, np.ndarray)

    def test_double_reverse_calculation(self):
        rho, Tv = calculate('rho', 'Tv', add_assumptions=('Tv equals T',),
                            **self.vars2)
        assert (rho == 1/Rd).all()
        assert isinstance(rho, np.ndarray)
        assert isinstance(Tv, np.ndarray)

    def test_T_from_Tv(self):
        assert calculate('T', Tv=1., add_assumptions=('Tv equals T',)) == 1.
        assert calculate('T', Tv=5., add_assumptions=('Tv equals T',)) == 5.

    def test_rv_from_qv(self):
        assert np.isclose(calculate('rv', qv=0.005), 0.005025125628140704)

    def test_qv_from_rv(self):
        assert np.isclose(calculate('qv', rv=0.005), 0.0049751243781094535)


class TestSolveValuesNearSkewT(unittest.TestCase):

    def setUp(self):
        self.quantities = {'p': 8.9e4, 'Tv': 4.5+273.15, 'theta': 14.+273.15,
                           'rv': 1e-3, 'Tlcl': -22.5+273.15,
                           'thetae': 17.+273.15, 'Tw': -2.5+273.15,
                           'Td': -18.5+273.15, 'plcl': 62500.,
                           }
        self.quantities['T'] = calculate('T', **self.quantities)
        self.quantities['rho'] = calculate('rho', **self.quantities)
        self.add_assumptions = ('bolton', 'unfrozen bulb')

    def _generator(self, quantity, tolerance):
        skew_T_value = self.quantities.pop(quantity)
        calculated_value, funcs = calculate(
            quantity, add_assumptions=self.add_assumptions,
            debug=True, **self.quantities)
        diff = abs(skew_T_value - calculated_value)
        if diff > tolerance:
            err_msg = ('Value {:.2f} is too far away from '
                       '{:.2f} for {}.'.format(
                           calculated_value, skew_T_value, quantity))
            err_msg += '\nfunctions used:\n'
            err_msg += '\n'.join([f.__name__ for f in funcs])
            raise AssertionError(err_msg)

    def tearDown(self):
        self.quantities = None

    def test_calculate_precursors(self):
        pass

    def test_calculate_p(self):
        self._generator('p', 10000.)

    def test_calculate_Tv(self):
        self._generator('Tv', 1.)

    def test_calculate_theta(self):
        self._generator('theta', 1.)

    def test_calculate_rv(self):
        self._generator('rv', 1e-3)

    def test_calculate_Tlcl(self):
        self._generator('Tlcl', 1.)

    def test_calculate_thetae(self):
        quantity = 'Tw'
        skew_T_value = self.quantities.pop(quantity)
        self.quantities.pop('Tlcl')  # let us calculate this ourselves
        calculated_value, funcs = calculate(
            quantity, add_assumptions=('bolton', 'unfrozen bulb'),
            debug=True,
            **self.quantities)
        diff = abs(skew_T_value - calculated_value)
        if diff > 1.:
            err_msg = ('Value {:.2f} is too far away from '
                       '{:.2f} for {}.'.format(
                           calculated_value, skew_T_value, quantity))
            err_msg += '\nfunctions used:\n'
            err_msg += '\n'.join([f.__name__ for f in funcs])
            raise AssertionError(err_msg)

    def test_calculate_Tw(self):
        quantity = 'Tw'
        skew_T_value = self.quantities.pop(quantity)
        calculated_value, funcs = calculate(
            quantity, add_assumptions=('bolton', 'unfrozen bulb'),
            debug=True,
            **self.quantities)
        diff = abs(skew_T_value - calculated_value)
        if diff > 1.:
            err_msg = ('Value {:.2f} is too far away from '
                       '{:.2f} for {}.'.format(
                           calculated_value, skew_T_value, quantity))
            err_msg += '\nfunctions used:\n'
            err_msg += '\n'.join([f.__name__ for f in funcs])
            raise AssertionError(err_msg)

#    def test_calculate_Td(self):
#        self._generator('Td', 1.)

    def test_calculate_plcl(self):
        self._generator('plcl', 10000.)


class TestSolveValuesNearSkewTAssumingLowMoisture(TestSolveValuesNearSkewT):

    def setUp(self):
        super(TestSolveValuesNearSkewTAssumingLowMoisture, self).setUp()
        self.add_assumptions = ('bolton', 'unfrozen bulb', 'low water vapor')


class TestSolveValuesNearSkewTVeryMoist(TestSolveValuesNearSkewT):

    def setUp(self):
        self.quantities = {'p': 8.9e4, 'Tv': 9.+273.15, 'theta': 18.4+273.15,
                           'rv': 6e-3, 'Tlcl': 3.8+273.15,
                           'thetae': 36.5+273.15,
                           'Tw': 6.5+273.15, 'Td': 4.8+273.15, 'plcl': 83500.,
                           }
        self.quantities['T'] = calculate('T', **self.quantities)
        self.quantities['rho'] = calculate('rho', **self.quantities)
        self.add_assumptions = ('bolton', 'unfrozen bulb')


class GetShortestSolutionTests(unittest.TestCase):

    def setUp(self):
        self.methods = {
            'x': {('a', 'b'): lambda a, b: a},
            'y': {('x',): lambda x: x},
            'z': {('x', 'y'): lambda x, y: x*y},
            'w': {('z', 'y'): lambda x, y: x*y},
        }

    def tearDown(self):
        self.methods = None

    def test_no_calculation(self):
        sol = _get_shortest_solution(('x',), ('x',), (), self.methods)
        assert isinstance(sol, tuple)
        assert len(sol) == 3
        assert len(sol[0]) == 0
        assert len(sol[1]) == 0
        assert len(sol[2]) == 0

    def test_simplest_calculation(self):
        sol = _get_shortest_solution(('y',), ('x',), (), self.methods)
        assert len(sol[0]) == 1

    def test_depth_2_calculation(self):
        _get_shortest_solution(('z',), ('a', 'b'), (), self.methods)

    def test_depth_3_calculation(self):
        _get_shortest_solution(('w',), ('a', 'b'), (), self.methods)


class EquationTests(unittest.TestCase):

    def _assert_accurate_values(self, func, in_values, out_values, tols):
        for i, args in enumerate(in_values):
            out_calc = func(*args)
            if abs(out_calc - out_values[i]) > tols[i]:
                raise AssertionError(
                    'Calculated value {} from inputs {} is more than {}'
                    'away from {}'.format(out_calc, args, tols[i],
                                          out_values[i]))

    def test_es_from_T_Bolton(self):
        func = equations.es_from_T_Bolton
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_es_from_T_Goff_Gratch(self):
        func = equations.es_from_T_Goff_Gratch
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_f_from_lat(self):
        func = equations.f_from_lat
        in_values = [(0.,), (45.,), (90.,)]
        out_values = [0., 1.028e-4, 1.454e-4]
        tols = [0.001e-4, 0.001e-4, 0.001e-4]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_Gammam_from_rvs_T(self):
        func = equations.Gammam_from_rvs_T
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_MSE_from_DSE_qv(self):
        func = equations.MSE_from_DSE_qv
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_omega_from_w_rho_hydrostatic(self):
        func = equations.omega_from_w_rho_hydrostatic
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_p_from_rho_Tv_ideal_gas(self):
        func = equations.p_from_rho_Tv_ideal_gas
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_plcl_from_p_T_Tlcl(self):
        func = equations.plcl_from_p_T_Tlcl
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_Phi_from_z(self):
        func = equations.Phi_from_z
        in_values = [(0.,), (30.,), (100.,)]
        out_values = [0., 294.3, 981.]
        tols = [0.01, 0.2, 1.]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qv_from_AH_rho(self):
        func = equations.qv_from_AH_rho
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qv_from_rv(self):
        func = equations.qv_from_rv
        in_values = [(0.,), (0.005,), (0.1,)]
        out_values = [0., 0.00498, 0.091]
        tols = [0.01, 0.00001, 0.001]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qv_from_rv_lwv(self):
        func = equations.qv_from_rv_lwv
        in_values = [(0.,), (0.005,), (0.1,)]
        out_values = [0., 0.005, 0.1]
        tols = [1e-8, 1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qv_from_p_e(self):
        func = equations.qv_from_p_e
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qv_from_p_e_lwv(self):
        func = equations.qv_from_p_e_lwv
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qvs_from_rvs(self):
        func = equations.qvs_from_rvs
        in_values = [(0.,), (0.005,), (0.1,)]
        out_values = [0., 0.00498, 0.091]
        tols = [0.01, 0.00001, 0.001]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qvs_from_rvs_lwv(self):
        func = equations.qvs_from_rvs_lwv
        in_values = [(0.,), (0.005,), (0.1,)]
        out_values = [0., 0.005, 0.1]
        tols = [1e-8, 1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qvs_from_p_es(self):
        func = equations.qvs_from_p_es
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qvs_from_p_es_lwv(self):
        func = equations.qvs_from_p_es_lwv
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qt_from_qi_qv_ql(self):
        func = equations.qt_from_qi_qv_ql
        in_values = [(0., 0., 0.), (0.005, 0.001, 0.004)]
        out_values = [0., 0.01]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qt_from_qv_ql(self):
        func = equations.qt_from_qv_ql
        in_values = [(0., 0.), (0.003, 0.002)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qt_from_qv(self):
        func = equations.qt_from_qv
        in_values = [(0.,), (0.005,), (0.1,)]
        out_values = [0., 0.005, 0.1]
        tols = [1e-8, 1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qt_from_qv_qi(self):
        func = equations.qt_from_qv_qi
        in_values = [(0., 0.), (0.003, 0.002)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qv_from_qt(self):
        func = equations.qv_from_qt
        in_values = [(0.,), (0.005,), (0.1,)]
        out_values = [0., 0.005, 0.1]
        tols = [1e-8, 1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qv_from_qt_ql_qi(self):
        func = equations.qv_from_qt_ql_qi
        in_values = [(0., 0., 0.), (0.01, 0.003, 0.002)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qv_from_qt_ql(self):
        func = equations.qv_from_qt_ql
        in_values = [(0., 0.), (0.01, 0.005)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qv_from_qt_qi(self):
        func = equations.qv_from_qt_qi
        in_values = [(0., 0.), (0.01, 0.005)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qi_from_qt_qv_ql(self):
        func = equations.qv_from_qt_ql_qi
        in_values = [(0., 0., 0.), (0.01, 0.003, 0.002)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_qi_from_qt_qv(self):
        func = equations.qi_from_qt_qv
        in_values = [(0., 0.), (0.01, 0.005)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_ql_from_qt_qv_qi(self):
        func = equations.qv_from_qt_ql_qi
        in_values = [(0., 0., 0.), (0.01, 0.003, 0.002)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_ql_from_qt_qv(self):
        func = equations.ql_from_qt_qv
        in_values = [(0., 0.), (0.01, 0.005)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_RH_from_rv_rvs(self):
        func = equations.RH_from_rv_rvs
        in_values = [(5., 100.), (1e-3, 2e-3)]
        out_values = [5., 50.]
        tols = [0.01, 0.01]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_RH_from_qv_qvs_lwv(self):
        func = equations.RH_from_qv_qvs_lwv
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rho_from_qv_AH(self):
        func = equations.rho_from_qv_AH
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rho_from_p_Tv_ideal_gas(self):
        func = equations.rho_from_p_Tv_ideal_gas
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rv_from_qv(self):
        func = equations.rv_from_qv
        in_values = [(0.,), (0.05,), (0.1,)]
        out_values = [0., 0.05263, 0.11111]
        tols = [1e-8, 0.0001, 0.0001]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rv_from_qv_lwv(self):
        func = equations.rv_from_qv_lwv
        in_values = [(0.,), (0.005,), (0.1,)]
        out_values = [0., 0.005, 0.1]
        tols = [1e-8, 1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rv_from_p_e(self):
        func = equations.rv_from_p_e
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rt_from_ri_rv_rl(self):
        func = equations.rt_from_ri_rv_rl
        in_values = [(0., 0., 0.), (0.005, 0.001, 0.004)]
        out_values = [0., 0.01]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rt_from_rv_rl(self):
        func = equations.rt_from_rv_rl
        in_values = [(0., 0.), (0.003, 0.002)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rt_from_rv(self):
        func = equations.rt_from_rv
        in_values = [(0.,), (0.005,), (0.1,)]
        out_values = [0., 0.005, 0.1]
        tols = [1e-8, 1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rt_from_rv_ri(self):
        func = equations.rt_from_rv_ri
        in_values = [(0., 0.), (0.003, 0.002)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rv_from_rt(self):
        func = equations.rv_from_rt
        in_values = [(0.,), (0.005,), (0.1,)]
        out_values = [0., 0.005, 0.1]
        tols = [1e-8, 1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rv_from_rt_rl_ri(self):
        func = equations.rv_from_rt_rl_ri
        in_values = [(0., 0., 0.), (0.01, 0.003, 0.002)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rv_from_rt_rl(self):
        func = equations.rv_from_rt_rl
        in_values = [(0., 0.), (0.01, 0.005)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rv_from_rt_ri(self):
        func = equations.rv_from_rt_ri
        in_values = [(0., 0.), (0.01, 0.005)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_ri_from_rt_rv_rl(self):
        func = equations.rv_from_rt_rl_ri
        in_values = [(0., 0., 0.), (0.01, 0.003, 0.002)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_ri_from_rt_rv(self):
        func = equations.ri_from_rt_rv
        in_values = [(0., 0.), (0.01, 0.005)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rl_from_rt_rv_ri(self):
        func = equations.rv_from_rt_rl_ri
        in_values = [(0., 0., 0.), (0.01, 0.003, 0.002)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rl_from_rt_rv(self):
        func = equations.rl_from_rt_rv
        in_values = [(0., 0.), (0.01, 0.005)]
        out_values = [0., 0.005]
        tols = [1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rvs_from_p_es(self):
        func = equations.rvs_from_p_es
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rvs_from_qvs(self):
        func = equations.rvs_from_qvs
        in_values = [(0.,), (0.05,), (0.1,)]
        out_values = [0., 0.05263, 0.11111]
        tols = [1e-8, 0.0001, 0.0001]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_rvs_from_qvs_lwv(self):
        func = equations.rvs_from_qvs_lwv
        in_values = [(0.,), (0.005,), (0.1,)]
        out_values = [0., 0.005, 0.1]
        tols = [1e-8, 1e-8, 1e-8]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_T_from_es_Bolton(self):
        func = equations.T_from_es_Bolton
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_Tlcl_from_T_RH(self):
        func = equations.Tlcl_from_T_RH
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_Tlcl_from_T_Td(self):
        func = equations.Tlcl_from_T_Td
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_Tlcl_from_T_e(self):
        func = equations.Tlcl_from_T_e
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_T_from_Tv_qv(self):
        func = equations.T_from_Tv_qv
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_Tv_from_T_qv(self):
        func = equations.Tv_from_T_qv
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_Tv_from_T_assuming_Tv_equals_T(self):
        func = equations.Tv_from_T_assuming_Tv_equals_T
        in_values = [(273.15,), (100.,), (300.,)]
        out_values = [273.15, 100., 300.]
        tols = [0.001, 0.001, 0.001]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_Tv_from_p_rho_ideal_gas(self):
        func = equations.Tv_from_p_rho_ideal_gas
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_Tw_from_T_RH_Stull(self):
        func = equations.Tw_from_T_RH_Stull
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_T_from_Tv_assuming_Tv_equals_T(self):
        func = equations.T_from_Tv_assuming_Tv_equals_T
        in_values = [(273.15,), (100.,), (300.,)]
        out_values = [273.15, 100., 300.]
        tols = [0.001, 0.001, 0.001]
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_theta_from_p_T(self):
        func = equations.theta_from_p_T
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_thetae_from_p_T_Tlcl_rv_Bolton(self):
        func = equations.thetae_from_p_T_Tlcl_rv_Bolton
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_thetae_from_p_e_T_RH_rv_rt(self):
        func = equations.thetae_from_p_e_T_RH_rv_rt
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_thetae_from_T_RH_rv_lwv(self):
        func = equations.thetae_from_T_RH_rv_lwv
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_thetaes_from_p_T_rvs_Bolton(self):
        func = equations.thetaes_from_p_T_rvs_Bolton
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_w_from_omega_rho_hydrostatic(self):
        func = equations.w_from_omega_rho_hydrostatic
        in_values = []
        out_values = []
        tols = []
        self._assert_accurate_values(func, in_values, out_values, tols)

    def test_z_from_Phi(self):
        func = equations.z_from_Phi
        in_values = [(0.,), (294.3,), (981.,)]
        out_values = [0., 30., 100.]
        tols = [1e-8, 0.1, 0.1]
        self._assert_accurate_values(func, in_values, out_values, tols)


if __name__ == '__main__':
    nose.run()
